; Computer Fantasies
; 256b by y0bi / wAMMA
; for Lovebyte 2024

; Řrřola disassembly + playing with the code: compile with NASM (needed for "$$" section start notation)

; SAVINGS
; - fs<->es
; - shld bp,si,...
; - read from PROG_STARTS only once, zip tables together
; - di(address) goes up, jns
; - I like black border + dithering instead of jitter
; - overlap opcode subroutines, cbw
; - ch is in bits (step=3, step inside bit read routine)... but max program size is smaller

; TODO
; - palette (hot + cold gradient?)
; - doubling: use "dup +" (2 opcodes) instead of "2 *" (3 opcodes)
; - try (X-64), (Y-64) instead of X, Y - would saves around 20 opcodes
; - instead of "dup", try "load(0..7)" with an immediate (like LIT): load(0)=dup, load(1)=over, or absolute
;   - reuse subexpressions: would save 7 opcodes in the last effect
;   - populate the stack with {time X Y} inputs and "load" them
;     - easier use of time
;     - would open two opcodes instead of "x" and "y" - suggestions:
;       - store(0..7)? would allow memory cells, maybe even persistent across frames; render to texture
;       - read neighboring pixels? would allow reading textures (could be also constant inputs)
;       - bit operations? "and", "xor"
;       - "sin" with a sine table?
;       - jump? ...but can't parallelize
;       - loops with constant number of iterations? 2+(0..7)
;       - keep both 1-opcode "dup" and "load(0..7)"
;       - a customized instruction for each program?

; RECIPES
;   make constants with math (-1 = 2/2 - 2)
;   x==-32768 ? 1 : 0  --> "x -32768/"
;   x>=0 ? 1 : 0       --> "x 8192+ 8192/"
;   x>=y ? t : f       -->  t + (f-t) * (x>=y)  -->  t + (f-t) * ((x-y + 8192) / 8192)
;   max(x,y)  -->  x>=y ? x : y
;   x%M       -->  dup M / M * -          ; truncating division; add a constant to emulate floor division

org 100h
            mov al,0x13
            int 0x10
            push 0xa000     ; es=a000
            pop es
            xor si,si       ; si: num_frame

      MAIN: push 0x3000     ; ds=3000: array segments are 3000 3800 4000 4800 5000 5800 6000 6800 ...
            pop ds

            shld bp,si,16-9 ; bp: prog_index = ((num_frame / 1024) % 4) * 2
            and bp,6
            mov cx,[bp+PROG_ADR_STARTS] ; ch: inst_bit_index (goes down), cl: prog address
            movzx bp,cl                 ; bp: prog address

 INST_LOOP: mov dx,ds
            mov fs,dx       ; fs = src segment

            call NEXT_INST  ; al = read next instruction
            cmp al,4        ; literal?

            mov bx,INST_OFFSETS
            ss xlatb
            mov bl,al       ; bh=0x1, bl=INST_OFFSETS[al] -> bx = inst subroutine

            jb INST_0123
            jne INST_567

            call NEXT_INST  ; if (inst==4) cl = read literal
            mov cl,al
            add cl,cl

  INST_567: add dh,0x10     ; if (bl>=4) dst=src+1  ; push a new value on the stack
 INST_0123: sub dh,0x08     ; else dst=src-1        ; pop stack
            mov ds,dx       ; ds = dst segment

            xor di,di       ; di = adr (0000..7ffe)

  ADR_LOOP: mov ax,[fs:di]  ; ax = src word, [di] = dst word
            mov dx,di       ; dx = adr (for INST_X and INST_Y)
            call bx         ; run inst subroutine
            mov [di],ax     ; [dst] = result

            test ch,ch
            jnz CH_ISNT_0

            push bx         ; on last inst, fill screen with lower byte of result

            mov [di],si     ; [dst] = num_frame
            mov bx,di
            mov bl,0
            shr bx,2        ; bx = y*64  -->  bx+di = screen_adr = y*320 + x

            shr al,4
            mov ah,al       ; al = 64 + (al>>4) + (al>>3 & 1)  ; dithering
            adc ax,0x4040   ; ah = 64 + (al>>4)
            mov [es:bx+di+(100-64)*320+32],ax

            pop bx

 CH_ISNT_0: inc di
            inc di
            jns ADR_LOOP    ; di=0x8000

            test ch,ch      ; end when inst_index=0
            jnz INST_LOOP

            inc si          ; num_frame++
            in al,0x60
            cmp al,1
            jnz MAIN        ; exit on esc

; Insts that pop from stack, result in ax
  INST_SUB: neg ax          ; { d a -> d-a }
  INST_ADD: add ax,[di]     ; { d a -> d+a }
            ret

  INST_MUL: imul word[di]   ; { d a -> d*a }
            ret

  INST_DIV: xchg ax,[di]    ; { d a -> d/a }
            cwd
            idiv word[di]
            ret

; Insts that push on stack
  INST_LIT: mov ax,2        ; { -> literal }  ; 0..7 -> 2 8 32 128 512 2048 8192 32768
            shl ax,cl
  INST_DUP: ret             ; { a -> a a }

    INST_Y: shr dx,7        ; { -> y }
    INST_X: shr dl,1        ; { -> x }
            xchg ax,dx
            cbw
            ret

; Read 3 bits from program, uses bp=program start, clobbers di
 NEXT_INST: mov di,cx
            shr di,8+3  ; di = bit >> 3
            mov cl,7
            and cl,ch   ; cl = bit & 7

            mov ax,[bp+di+0x100]
            shr ax,cl
            and al,7    ; read 3 bits

            sub ch,3    ; next inst_index
            ret

INST_OFFSETS:  db INST_ADD-$$, INST_SUB-$$, INST_MUL-$$, INST_DIV-$$
               db INST_LIT-$$, INST_X-$$, INST_Y-$$, INST_DUP-$$
PROG_ADR_STARTS:
               db PROG0-$$, 6*3    ; address (in bytes), first inst_index (in bits)
               db PROG1-$$, 30*3
               db PROG2-$$, 24*3
               db PROG3-$$, 62*3

               ; Opcodes:         0 1 2  3   4   5    6    7
               ; 0+ 1- 2* 3/ 4lit[2 8 32 128 512 2048 8192 -32768] 5x 6y 7dup(.)

       PROG0:  db 0x80,0x6C,0x15
               ;   t x 8192 y * + +
               ;   color = t + x + y*8192

       PROG1:  db 0xC0,0x46,0x65,0xE0,0x58,0x90,0xD1,0x33,0x70,0xB4,0x29,0x11
               ;   t 32* 8192 y 128 2/ - .* - 512 + x 128 2/ - * 128 / + +
               ;   t*32 8192 y-64 ² - 512+ x-64 * 128/ + +
               ;   <- t*32 + (8704-(y-64)²) * (x-64) / 128

       PROG2:  db 0xC8,0x00,0x42,0x0F,0x2E,0x5A,0x0F,0x2E,0xF6,0x04
               ;   t -32768 x 64- .* y 64- .* + 2+ / - +
               ;   t -32768 x-64 ² y-64 ² + 2+ / - +
               ;   <- t - -32768 / ((x-64)² + (y-64)² + 2)

       PROG3:  db 0x80,0x9E,0x81,0x62,0x24,0x4E,0x3C,0xC6,0x8D,0xCA,0xC0,0xB1,0x03,0x68,0x53,0xCF,0x40,0x31,0x12,0x27,0x1E,0x03,0x68,0x13
               ; t 8192 y 2+ / + .32/32*- 16- .* 8192 y 2+ / x 64- * 128/ 128+ .32/32*- 16- .* + +
               ; t 8192 y+2 / + 32% 16- ² 8192 y+2 / x-64 * 128/ 128+ 32% 16- ² + +
               ;   v = (t + 8192/(y+2)) % 32
               ;   u = (8192/(y+2) * (x-64) / 128 + 128) % 32
               ;   <- (u-16)² + (v-16)²
