; Computer Fantasies
; 256b by y0bi / wAMMA
; for Lovebyte 2024

; Řrřola disassembly + playing with the code: compile with NASM (needed for "$$" section start notation)

; SAVINGS v1
; - fs<->es
; - shld bp,si,...
; - read from PROG_STARTS only once, zip tables together
; - di(address) goes up, jns
; - I like black border + dithering instead of jitter
; - overlap opcode subroutines, cbw
; - ch is in bits (step=3, step inside bit read routine)... but max program size is smaller

; SAVINGS v2
; - [es:0] = top left pixel on the screen
; - do "shr dl,1" before "call bx", then "xchg dl,dh"
; - use x-64, y-64 instead of x, y
; - ax=si at MAIN (from last frame when it was "pushed" onto the stack)
; - read speculatively 8 bits of program, split them into 3+5, inline read call (but bit index needs to go up)
; - use 4-bit literals (1<<i): saves 2 bytes in the interpreter, no need for (dup +) or (2 *) for some constants

; TODO
; - program length at [bp-1] is a whole byte and limiting: a "ret" instruction would be better, save 5 bits
;   - pair with "call(imm->program number)"? can use the regular stack for return values
;   - nice Forth-like program factoring
;   - loops with constant number of iterations (call multiple times)
; - inst arguments with different bit widths: many new possibilites (x,y -> single inst with 1-bit immediate?)
; - palette (hot + cold gradient?)
; - instead of "dup", try "load" with an 1-bit or 2-bit immediate: load(0)=dup, load(1)=over
;   - easier use of time
;   - push x and y onto the initial stack too? {x y t} would be nice
; - other opcode suggesions:
;   - store(0..7)? would allow persistent memory cells across frames
;   - bit operations? and, xor
;   - "*/" muldiv (16*16->32bit mul, than 32/16->16bit div)
;   - trig with tables? sin, atan2

; RECIPES
;   make constants with math (7: 8 1-, 21: 16 4 1++)
;   drop0              --> dup -
;   drop               --> dup - +
;   nip                --> over - +
;   0                  --> dup dup -
;   x==-32768 ? 1 : 0  --> x -32768/
;   x>=0 ? 1 : 0       --> x 8192+ 8192/
;   x>=y ? t : f       -->  t + (f-t) * (x>=y)  -->  t + (f-t) * ((x-y + 8192) / 8192)
;   max(x,y)           -->  x>=y ? x : y
;   x%M                -->  dup M / M * -          ; truncating division; add a constant to emulate floor division

org 100h
            mov al,0x13
            int 0x10
            push 0xa000 + ((100-64)*320 + 32) / 16  ; es=a2d2, centered 256*128 rectangle
            pop es
            xor si,si       ; si: num_frame (-2 bytes: can omit if we don't mind starting at 0x0100)

      MAIN: push 0x3000     ; ds=3000: array segments are 3000 3800 4000 4800 5000 5800 6000 6800 ...
            pop ds
            xor cx,cx       ; ch = bit index (goes up)

            mov bx,PROG_OFFSETS ; bh=1 always
            mov al,ah       ; ax = si = frame_num (from last frame)
            and al,3        ; al: prog_index = (num_frame / 256) % 4
            ss xlat         ; al: prog_offset = [ss:bx + prog_index]
            mov ah,1
            xchg ax,bp      ; bp = prog_address

 INST_LOOP: mov di,cx
            shr di,8+3      ; di = bit >> 3
            mov cl,7
            and cl,ch       ; cl = bit & 7
            add ch,3        ; advance by 3 bits (inst size)

            mov ax,[bp+di]
            shr ax,cl       ; ax = peek next 9..16 bits from program (we'll use only 8)
            aam 8           ; al = low 3 bits (next inst), ah = high 5 bits (optional argument)

            mov dx,ds
            mov fs,dx       ; fs = src segment

            cmp al,4        ; literal?

            mov bl,INST_OFFSETS-$$
            ss xlat
            mov bl,al       ; bh=1, bl=INST_OFFSETS[al] -> bx = inst subroutine
            mov cl,ah       ; cl = literal argument (5 bits but only 4 are used)

            jb INST_0123
            jne INST_567
            add ch,4        ; advance by 4 bits (literal arg size)
  INST_567: add dh,0x10     ; if (bl>=4) dst=src+1  ; inst pushes a new value on the stack
 INST_0123: sub dh,0x08     ; else dst=src-1        ; inst pops stack
            mov ds,dx       ; ds = dst segment

            xor di,di       ; di = adr (0000..7ffe)

  ADR_LOOP: mov ax,[fs:di]  ; ax = src word, [di] = dst word
            mov dx,di
            shr dl,1        ; dh = y, dl = x, cl = argument
            call bx         ; run inst subroutine

            cmp ch,[bp-1]   ; last instruction?
            jb NOT_LAST     ; cf=1, will also be tested later

            push bx         ; on last inst, fill screen with lower byte of result

;            mov bl,dh      ; (-2 bytes): if the last instruction wasn't INST_X or INST_Y, then dh=y
;            shl bx,6       ; (also need to subtract 0x10<<6 from es because bh=1 here)
            mov bx,di
            mov bl,0
            shr bx,2        ; bx = y*64  -->  bx+di = screen_adr = y*320 + x

            shr al,4
            mov ah,al       ; al = 63 + (al>>4) + (al>>3 & 1)  ; dithering
            adc ax,0x3f3f   ; ah = 63 + (al>>4)                ; cf=0, will be tested later
            mov [es:bx+di],ax ; es is offset so that the rectangle is centered on screen

            pop bx
            mov ax,si       ; ax = num_frame; start next frame with T pushed on the stack

  NOT_LAST: mov [di],ax     ; [di] = store result to dst word
            inc di
            inc di
            jns ADR_LOOP    ; di=0x8000

            jb INST_LOOP    ; use carry flag from "cmp ch,[bp-1]"

            inc si          ; num_frame++
            in al,0x60
            cmp al,1
            jnz MAIN        ; exit on esc

; Insts that pop from stack, result in ax
  INST_SUB: neg ax          ; { d a -> d-a }
  INST_ADD: add ax,[di]     ; { d a -> d+a }
            ret

  INST_MUL: imul word[di]   ; { d a -> d*a }
            ret

  INST_DIV: xchg ax,[di]    ; { d a -> d/a }
            cwd
            idiv word[di]
            ret

; Insts that push on stack
  INST_LIT: mov ax,1        ; { -> literal }  ; 0..15 -> 1 2 4 8 ... 32768
            rol ax,cl       ; use 4 bits of argument
  INST_DUP: ret             ; { a -> a a }

    INST_Y: xchg dl,dh      ; { -> y }
    INST_X: xchg ax,dx      ; { -> x }
            sub al,64       ; center
            cbw
            ret

 INST_OFFSETS: db INST_ADD-$$, INST_SUB-$$, INST_MUL-$$, INST_DIV-$$
               db INST_LIT-$$, INST_X-$$, INST_Y-$$, INST_DUP-$$

 PROG_OFFSETS: db PROG0-$$, PROG1-$$, PROG2-$$, PROG3-$$  ; si=0x100 on init - start with PROG0

               ; Opcodes:
               ; 0+ 1- 2* 3/ 4lit[1 << 4-bit immediate] 5x 6y 7dup(.)

               db 18  ; program length in bits
       PROG0:  db 10_010_101b,0_010_111_1b,000000_00b
               ; 101 010 110 111 010 000
               ; x   *   y   .   *   +
               ; (t) x * y ² +
               ; color = t*x + y²    ; <- different because x,y is now x-64,y-64

               db 58
       PROG1:  db 0_0101_100b,101_100_01b,1001_100_1b,11_110_000b,1_001_010_1b,100_010_10b,0_011_0111b,000000_00b
               ; 100 0101 010 100 1101 100 1001 000 110 111 010 001 101 010 100 0111 011 000
               ; 32       *   8192     512      +   y   .   *   -   x   *   128      /   +
               ; (t) 32* 8192+512 y ² - x * 128/ +
               ; color = t*32 + (8704-y²)*x/128

               db 50
       PROG2:  db 11111100b,01011110b,10000111b,00101111b,00110000b,10110000b,00000000b
               ; 100 1111 101 111 010 111 000 110 111 010 000 100 0001 000 011 001
               ; -32768   x   .   *   .   +   y   .   *   +   2        +   /   -
               ; (t) -32768 x ² .+ y ² + 2+ / -
               ; color = t + 32768 / (2x² + y² + 2)      ; <- modified to use 2x²

               db 182
       PROG3:  db 01101100b,00110011b,01000000b,11000011b,01110000b,11010110b,01011000b,00001010b,00101001b,00010111b,
               db 11011011b,00001100b,11010000b,10110000b,10001010b,00110111b,00011110b,11001110b,00011010b,01001011b,
               db 00100001b,11100101b,00000010b,
               ; 100 1101 110 100 0001 000 100 0110 000 011  000
               ; 8192     y   2        +   64       +   /    +

               ; 111 100 0101 011 100 0101 010 001 100 0100 001 111 010
               ; .   32       /   32       *   -   16       -   .   *

               ; 100 1101 110 100 0001 000 100 0110 000 011  101 010 100 0111 011 100 0111 000
               ; 8192     y   2        +   64       +   /    x   *   128      /   128      +

               ; 111 100 0101 011 100 0101 010 001 100 0100 001 111 010  000
               ; .   32       /   32       *   -   16       -   .   *    +

               ; 8192 y+66 /  +  32% 16- .*  8192 y+66 /  x * 128/ 128+  32% 16- .*  +
               ; v = (8192/(y+66) + t) % 32 - 16
               ; u = (8192/(y+66) * x / 128 + 128) % 32 - 16
               ; color = u² + v²
