; Computer Fantasies
; 256b by y0bi / wAMMA
; for Lovebyte 2024

; Řrřola disassembly + playing with the code: compile with NASM (needed for "$$" section start notation)

; v1 - optimizations
; - fs<->es
; - shld bp,si,...
; - read from PROG_STARTS only once, zip tables together
; - di(address) goes up, jns
; - I like black border + dithering instead of jitter
; - overlap opcode subroutines, cbw
; - ch is in bits (step=3, step inside bit read routine)... but max program size is smaller

; v2 - more optimizations
; - [es:0] = top left pixel on the screen
; - do "shr dl,1" before "call bx", then "xchg dl,dh"
; - use x-64, y-64 instead of x, y
; - ax=si at MAIN (from last frame when it was "pushed" onto the stack)
; - read speculatively 8 bits of program, split them into 3+5, inline read call (but bit index needs to go up)
; - use 4-bit literals (1<<i): saves 2 bytes in the interpreter, no need for (dup +) or (2 *) for some constants

; v3 - xy (+9 bytes), load instead of dup (+16 bytes)
; - "x", "y" joined into one instruction "xy": +9 bytes, +1 additional bit each time they're used
; - "load" with a 3-bit immediate: load(0)=dup, load(1)=over, ..., load(deep)=pop
;   With a corresponding "store" instruction we could have variables (even persistent across frames)
;   - emulate a second stack (push, pop); store(0)=drop, store(1)=nip, ..., store(deep)=push
;   - push x and y onto the initial stack too? {x y t} would be nice

; v4 - subroutines (+23 bytes)
; - "x", "y", "call" can't be the last instruction before return
; - "call" with a 3-bit immediate: calls another program, uses the regular x86 stack
; - call one program multiple times: loops with constant number of iterations

; TODO
; - program length at [bp-1] is a whole byte and limiting: a "ret" instruction would be better, save 5 bits
;   - loops with constant number of iterations ()
; - try arguments with different bit widths
; - palette (hot + cold gradient?)
; - other opcode suggesions:
;   - bit operations? and, xor
;   - "*/" muldiv (16*16->32bit mul, than 32/16->16bit div)
;   - trig with tables? sin, atan2

; RECIPES
;   make constants with math (7: 8 1-, 21: 16 4 1++)
;   drop 0             --> dup -
;   drop               --> dup - +
;   nip                --> over - +
;   0                  --> dup dup -
;   x==-32768 ? 1 : 0  --> x -32768/
;   x>=0 ? 1 : 0       --> x 8192+ 8192/
;   x>=y ? t : f       -->  t + (f-t) * (x>=y)  -->  t + (f-t) * ((x-y + 8192) / 8192)
;   max(x,y)           -->  x>=y ? x : y
;   x%M                -->  dup M / M * -          ; truncating division; add a constant to emulate floor division

org 100h
            mov al,0x13
            int 0x10
            push 0xa000 + ((100-64)*320 + 32) / 16 - (0x10 << 6) ; es=a2d2, centered 256*128 rectangle
            pop es
            xor si,si       ; si: num_frame (-2 bytes: can omit if we don't mind starting at 0x0100)

      MAIN: push 0x7800     ; ds=7800: array segments are 7800 7000 6800 6000 ...
            pop ds          ; can access also 8000 8800 9000 9800 as scratch segments
            mov al,3

 INST_CALL: mov bx,PROG_OFFSETS ; bh=1 always, ax = si = frame_num (from last frame) or ax = called program
            xor cx,cx       ; ch = bit index (goes up)
            and al,ah       ; al: prog_index = (num_frame / 256) % 4
            ss xlat         ; al: prog_offset = [ss:bx + prog_index]
            mov ah,1
            xchg ax,bp      ; bp = prog_address
; TODO no need for both ch and bp to address the program

 INST_LOOP: mov di,cx
            shr di,8+3      ; di = bit >> 3
            mov cl,7
            and cl,ch       ; cl = bit & 7
            add ch,3        ; advance by 3 bits (inst size)

            mov ax,[bp+di]
            shr ax,cl       ; ax = peek at next 9..16 bits from program (we'll use only 8)
            aam 8           ; al = low 3 bits (next inst), ah = high 5 bits (optional argument)
            mov cl,ah       ; cl = argument

            cmp al,7        ; call?
            jne INST_NOT7
            add ch,3        ; advance by 3 bits (inst size)
            pusha
            jmp INST_CALL

 INST_NOT7: cmp al,2        ; set flags

            mov bl,INST_OFFSETS-$$
            ss xlat
            mov bl,al       ; bh=1, bl=INST_OFFSETS[al] -> bx = inst subroutine

            mov ax,ds
            mov fs,ax

            ja INST_3456
            je INST_2
            jpe INST_1
    INST_0: add ch,3        ; inst_lit: advance by 4 bits
    INST_2: inc ch          ; inst_xy: advance by 1 bit
    INST_1:                 ; inst_dup: don't advance

            sub ah,0x10     ; if (bl>=4) dst=src+1  ; inst pushes a new value on the stack
 INST_3456: add ah,0x08     ; else dst=src-1        ; inst pops stack
            mov ds,ax       ; ds = dst segment

            xor di,di       ; di = adr (0000..7ffe)

  ADR_LOOP: mov ax,[fs:di]  ; ax = src word, [di] = dst word
            mov dx,di
            shr dl,1        ; dh = y, dl = x, cl = argument
            call bx         ; run inst subroutine

            cmp ch,[bp-1]   ; last instruction?
            jb NOT_LAST     ; cf=1
            cmp sp,-2
            clc             ; cf=0
            jne NOT_LAST

            push bx

            mov bl,dh       ; (bh=1 here so need to subtract 0x10<<6 from es)
            shl bx,6        ; bx = y*64  -->  bx+di = screen_adr = y*320 + x

            shr al,4
            mov ah,al       ; al = 63 + (al>>4) + (al>>3 & 1)  ; dithering
            adc ax,0x3f3f   ; ah = 63 + (al>>4)                ; cf=0, will be tested later

            mov [es:bx+di],ax ; es is offset so that the rectangle is centered on screen
            mov ax,si       ; ax = num_frame; start next frame with T pushed on the stack

            pop bx

  NOT_LAST: mov [di],ax     ; [di] = store result to dst word
            inc di
            inc di
            jns ADR_LOOP    ; di=0x8000

            jb INST_LOOP

            cmp sp,-2       ; top level?
            je TOPLV
            popa
            jmp INST_LOOP

     TOPLV: inc si          ; num_frame++
            in al,0x60
            cmp al,1
            jnz MAIN        ; exit on esc

; Insts that pop from stack, result in ax
  INST_SUB: neg ax          ; { d a -> d-a }
  INST_ADD: add ax,[di]     ; { d a -> d+a }
            ret

  INST_MUL: imul word[di]   ; { d a -> d*a }
            ret

  INST_DIV: xchg ax,[di]    ; { d a -> d/a }
            cwd
            idiv word[di]
            ret

; Insts that push on stack
  INST_LIT: mov ax,1        ; { -> literal }  ; 0..15 -> 1 2 4 8 ... 32768
            rol ax,cl       ; use 4 bits of 5-bit argument
  INST_DUP: ret             ; { a -> a a }

   INST_XY: test cl,1
            jz INST_X
            xchg dl,dh      ; { -> y }
    INST_X: xchg ax,dx      ; { -> x }
            sub al,64       ; center
            cbw
            ret

 INST_OFFSETS: db INST_LIT-$$, INST_DUP-$$, INST_XY-$$
               db INST_ADD-$$, INST_SUB-$$, INST_MUL-$$, INST_DIV-$$
               ; INST_CALL doesn't have a subroutine

 PROG_OFFSETS: db PROG0-$$, PROG1-$$, PROG2-$$, PROG3-$$
               db PROG4-$$, PROG5-$$ ; these are only used as called procedures

               ; Opcodes:
               ; 0lit[1 << 4-bit immediate] 1dup(.) 2xy[1-bit: 0->x, 1->y] 3+ 4- 5* 6/ 7call[3-bit]

               db 20  ; program length in bits
       PROG0:  db 01010010b,01001101b,0111b
               ; 010 0 101 010 1 001 101 011
               ; x     *   y     .   *   +
               ; (t) x * y ² +
               ; color = t*x + y²    ; <- different because x,y is now x-64,y-64

               db 60
       PROG1:  db 10101000b,10100010b,10010001b,11010011b,10010100b,01010010b,10011100b,0111b
               ; 000 0101 101 000 1101 000 1001 011 010 1 001 101 100 010 0 101 000 0111 110 011
               ; 32       *   8192     512      +   y     .   *   -   x     *   128      /   +
               ; (t) 32* 8192+512 y ² - x * 128/ +
               ; color = t*32 + (8704-y²)*x/128

               db 52
       PROG2:  db 01111000b,01001001b,00110011b,01001101b,10000111b,10011000b,1001b
               ; 000 1111 010 0 001 101 001 011 010 1 001 101 011 000 0001 011 110 100
               ; -32768   x     .   *   .   +   y     .   *   +   2        +   /   -
               ; (t) -32768 x ² .+ y ² + 2+ / -
               ; color = t + 32768 / (2x² + y² + 2)      ; <- modified to use 2x²

               db 57
       PROG3:  db 11100111b,11011110b,01010011b,10001010b,00110011b,01101110b,11101111b,0b
               ; 111 100 011 111 101 111 100  010 0 101 000 0111 110 000 0111 011 111 101 011
               ; call4   +   call5   call4    x     *   128      /   128      +   call5   +

               db 34
       PROG4:  db 01101000b,01000101b,00001100b,00110110b,11b
               ; 000 1101 010 1 000 0001 011 000 0110 011 110
               ; 8192     y     2        +   64       +   /
               ; { -> 8192/(y+66) }

               db 42
       PROG5:  db 01000001b,00011001b,01010101b,10000010b,10011000b,10b
               ; 001 000 0101 110 000 0101 101 100 000 0100 100 001 101
               ; .   32       /   32       *   -   16       -   .   *
               ; { a -> (a%32 - 16)² }
